/*
 * File:        wx_gdi.cc
 * Purpose:        GDI (Graphics Device Interface) objects and functions
 * Author:        Julian Smart
 * Created:        1993
 * Updated:        August 1994
 * RCS_ID:      $Id: wx_gdi.cc,v 1.1 1994/08/14 21:59:17 edz Exp $
 * Copyright:        (c) 1993, AIAI, University of Edinburgh
 */

/* static const char sccsid[] = "%W% %G%"; */

#pragma hdrstop
#include <iostream.h>
#include <stdio.h>
#include "common.h"
#include "wx_utils.h"
#include "wx_win.h"
#include "wx_gdi.h"
#include "wx_main.h"
#include "wx_rsrc.h"

#if USE_XPM_IN_OS2
#define FOR_OS2 1
#include "..\..\contrib\wxxpm\libxpm.34b\lib\xpm34.h"
#endif

#if USE_IMAGE_LOADING_IN_OS2
#include "..\..\utils\dib\dib.h"
#endif

#if USE_RESOURCE_LOADING_IN_OS2
#include "..\..\utils\rcparser\src\wxcurico.h"
#endif

Bool wxMakeBitmapAndPalette(PBITMAPINFOHEADER lpInfo,
                        HPAL * phPal, HBITMAP * phBitmap);

//
// Local Id generation for fonts and bitmaps
// Generates font reusable id's between 1 and 255
//
static long wxFreeLcids[8] = {0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
			      0xffffffff, 0xffffffff, 0xffffffff, 0x8fffffff};

static long 
wxNewLcid(void)
{
    int w, lcid = -1;
    for (w = 0; w < 8; w++) {
	if (wxFreeLcids[w]) {
	    int bit = ffs(wxFreeLcids[w]);
	    wxFreeLcids[w] &= ~(1 << (bit-1));
	    lcid = bit + w * 32;
	    break;
	}
    }
    TRACE("wxNewLcid() => %d", lcid);
    if (lcid)
	return lcid;

    /*
     * Out of LCID's. Panic
     */
    assert(0);
    return 0;				// To keep compiler happy
}

static void
wxFreeLcid(long lcid)
{
    TRACE("wxFreeLcid(lcid=%d)", lcid);
    int w = (lcid - 1) / 32;
    int bit = (lcid - 1) % 32;
    assert((wxFreeLcids[w] & (1 << bit)) == 0);
    wxFreeLcids[w] |= (1 << bit);
}

// #define DEBUG_CREATE
// #define TRACE_GDI

IMPLEMENT_DYNAMIC_CLASS(wxFont, wxObject)

wxFont::wxFont(void)
{
    TRACE("wxFont::wxFont()");
    point_size = 0;
    temporary = TRUE;
    cfont = 0;
    cxChar = cyChar = 0;
    wxTheFontList->Append(this);
}

/* Constructor for a font. Note that the real construction is done
 * in wxDC::SetFont, when information is available about scaling etc.
 */
wxFont::wxFont(int PointSize, int Family, int Style, int Weight, Bool Underlined, const char *Face):
    wxbFont(PointSize, Family, Style, Weight, Underlined)
{
    TRACE("wxFont::wxFont(PointSize=%d, Family='%d', Style=%d, Weight=%d, Underlined=%d, Face='%s'", PointSize, Family, Style, Weight, Underlined, Face);
    point_size = 0;
    temporary = TRUE;
    cfont = 0;
    cxChar = cyChar = 0;
    wxTheFontList->Append(this);

    Create(PointSize, Family, Style, Weight, Underlined, Face);
}

/*
 * Constructor for font given an OS/2-style font descriptor.
 *
 * OS/2 type font descriptor:
 *	<fontsize> . <Font name>> { . Outline } { . Strikeout } { . Underscore }
 */
wxFont::wxFont(const char *fontdesc)
{
    TRACE("wxFont::wxFont(fontdesc='%s')", fontdesc);
    int size;
    Bool outline = FALSE;
    Bool strikeout = FALSE;
    Bool underscore = FALSE;
    char *cp, *cp1;
    char facename[80];

    point_size = 0;
    temporary = TRUE;
    cfont = 0;
    cxChar = cyChar = 0;
    wxTheFontList->Append(this);

    cp = (char *)fontdesc;
    // Collect font size
    while (*cp >= '0' && *cp <= '9') {
	size = size * 10 + (*cp - '0');
	cp++;
    }
    if (size != 0 && *cp == '.') {
	// Collect facename
	cp1 = facename;
	while (++*cp && *cp != '.')
	    *cp1++ = *cp;
	*cp1 = '\0';

	if (strncmp(cp, ".Outline", 8) == 0) {
	    outline = TRUE;
	    cp += 8;
	}
	if (strncmp(cp, ".Strikeout", 10) == 0) {
	    strikeout = TRUE;
	    cp += 10;
	}
	if (strncmp(cp, ".Underscore", 11) == 0) {
	    underscore = TRUE;
	    cp += 11;
	}
	if (*cp == '\0') {
	    // Collected a legal font name.
	    // Convert to wxWindows type parameters
	    int fid, family, style, weight;
	    char *familyname;
	    if (wxTheFontNameDirectory.GetPsFontData(facename, &fid, &family, &familyname, &style, &weight))
		Create(size, family, style, weight, underscore, familyname);
	}
    }
}    


Bool wxFont::Create(int PointSize, int Family, int Style, int Weight, Bool Underlined, const char *Face)
{
    TRACE("wxFont::Create(PointSize=%d, Family='%d', Style=%d, Weight=%d, Underlined=%d, Face='%s')", PointSize, Family, Style, Weight, Underlined, Face);
#if 1
    if (Face)
	faceName = copystring(Face);
    else
	faceName = NULL;

    if (Face) {
	fontid = wxTheFontNameDirectory.FindOrCreateFontId(Face, Family);
	family = wxTheFontNameDirectory.GetFamily(fontid);
    } else {
	fontid = Family;
	family = wxTheFontNameDirectory.GetFamily(Family);
    }
    TRACE("Fontid=%d", fontid);
    style = Style;
    weight = Weight;
    point_size = PointSize;
    underlined = Underlined;

    wxTheFontList->Append (this);

#else
    family = Family;
    style = Style;
    weight = Weight;
    point_size = PointSize;
    underlined = Underlined;
    if (Face)
	faceName = copystring(Face);
    else
	faceName = NULL;

    temporary = FALSE;
    cfont = 0;
    cxChar = cyChar = 0;

#endif
    return TRUE;
}

wxFont::~wxFont()
{
    TRACE("wxFont::~wxFont() this=%#x", this);
    if (cfont) {
// OOps
//	GpiDeleteSetId(cfont);
	wxFreeLcid(cfont);
	cfont = 0;
    }
    wxTheFontList->DeleteObject(this);
}

void wxFont::BuildInternalFont(HPS ps)
{
    TRACE("wxFont::BuildInternalFont(ps=%#x)", (int)ps);

    if (!cfont) {
	BYTE ff_italic;
	int ff_weight = 0;
	int ff_family = 0;
	char *ff_face = NULL;
	USHORT fsSelection = 0;

	FATTRS fat;
	
	if (underlined)
	    fsSelection |= FATTR_SEL_UNDERSCORE;

	fat.usRecordLength = sizeof(FATTRS);
	fat.fsSelection = fsSelection;
	fat.lMatch = 0L;
	fat.idRegistry = 0;
	fat.usCodePage = 0;
	fat.lMaxBaselineExt = 0;
	fat.lAveCharWidth = 0;
	fat.fsType = FATTR_FONTUSE_OUTLINE | FATTR_FONTUSE_TRANSFORMABLE;
	fat.fsFontUse = 0;
#if 0
	strcpy(fat.szFacename, GetOS2FaceName());
#else

TRACE("wxFont::BuildInternalFont: Fontid=%d", family);
TRACE("wxFont::BuildInternalFont: GetScreenName() => '%s'", wxTheFontNameDirectory.GetScreenName(family, weight, style));
TRACE("wxFont::BuildInternalFont: GetPostScriptName() => '%s'", wxTheFontNameDirectory.GetPostScriptName(family, weight, style));
TRACE("wxFont::BuildInternalFont: GetAFMName() => '%s'", wxTheFontNameDirectory.GetAFMName(family, weight, style));
TRACE("wxFont::BuildInternalFont: GetFontName() => '%s'", wxTheFontNameDirectory.GetFontName(family));
TRACE("wxFont::BuildInternalFont: GetFamily() => %d", wxTheFontNameDirectory.GetFamily(family));

#if defined(wx_x) || (defined(wx_msw) && USE_PORTABLE_FONTS_IN_MSW) || defined(wx_os2)
	ff_face = wxTheFontNameDirectory.GetScreenName(family, weight, style);
	if (!*ff_face)
	    ff_face = NULL;
#else
	if (faceName)
	    ff_face = faceName;
#endif
	TRACE("facename = '%s'", ff_face);
	strcpy(fat.szFacename, ff_face);

#endif
 
	cfont = wxNewLcid();
	TRACE("wxFont::BuildInternalFont: Creating font #%d Facename=%s, ", cfont, fat.szFacename);
	LONG lMatch = GpiCreateLogFont(ps, NULL, cfont, &fat);
	TRACE("GpiCreateLogFont returned %ld", lMatch);

	FONTMETRICS fm;
	GpiQueryFontMetrics(ps, sizeof(FONTMETRICS), &fm);
	cxChar = fm.lAveCharWidth;
	cyChar = fm.lMaxAscender /* + fm.lMaxDescender*/ /* + fm.lExternalLeading */;
	TRACE("MaxAscender=%d, MaxDescender=%d", (int)fm.lMaxAscender, (int)fm.lMaxDescender);
    }
}

/*
 * Return the face name of a font.
 */
const char *wxFont::GetOS2FaceName()
{
    switch (family) {
    case wxROMAN:
	if (style == wxITALIC || style == wxSLANT)
	    if (weight == wxBOLD)
		return "Times New Roman Bold Italic";
	    else
		return "Times New Roman Italic";
	else if (weight == wxBOLD)
	    return "Times New Roman Bold";
	else
	    return "Times New Roman";
	break;

    case wxTELETYPE:
    case wxMODERN:
	if (style == wxITALIC || style == wxSLANT)
	    if (weight == wxBOLD)
		return "Courier Bold Italic";
	    else
		return "Courier Italic";
	else if (weight == wxBOLD)
	    return "Courier Bold";
	else
	    return "Courier";
	break;
	  
    case wxSWISS:
	if (style == wxITALIC || style == wxSLANT)
	    if (weight == wxBOLD)
		return "Helvetica Bold Italic";
	    else
		return "Helvetica Italic";
	else if (weight == wxBOLD)
	    return "Helvetica Bold";
	else
	    return "Helvetica";
	break;
	
    case wxDEFAULT:
    default:
	return "Helvetica"; 
    }
}


LONG wxFont::GetInternalFont(HPS ps)
{
    BuildInternalFont(ps) ;
    return cfont;
}

/* 
 * Return Name.Size for a font
 */
char *wxFont::GetNameSize(void)
{
    static char scratch[80];

    sprintf(scratch, "%d.%s", GetPointSize(), GetPostScriptName());
    TRACE("wxFont::GetNameSize() => '%s'", scratch);
    return scratch;
}

char *wxFont::GetPostScriptName(void)
{
    TRACE("wxbFont::GetPostScriptName()");
    return wxTheFontNameDirectory.GetPostScriptName(fontid, weight, style); 
}

/*
 * Colour map
 *
 */

IMPLEMENT_DYNAMIC_CLASS(wxColourMap, wxObject)

wxColourMap::wxColourMap(void)
{
    os2_palette = 0;
}

wxColourMap::wxColourMap(const int n, const unsigned char *red, const unsigned char *green, const unsigned char *blue)
{
    Create(n, red, green, blue);
}

wxColourMap::~wxColourMap(void)
{
    TRACE("wxColourMap::~wxColourMap() this=%#x", this);
    if (os2_palette) {
#if 0
	GpiDeletePalette(os2_palette);
#endif    
    }
}

Bool wxColourMap::Create(const int n, const unsigned char *red, const unsigned char *green, const unsigned char *blue)
{
    if (os2_palette)
	return FALSE;
#if 0    
    NPLOGPALETTE npPal = (NPLOGPALETTE)LocalAlloc(LMEM_FIXED, sizeof(LOGPALETTE) + 
						  (WORD)n * sizeof(PALETTEENTRY));
    if (!npPal)
	return(FALSE);

    npPal->palVersion = 0x300;
    npPal->palNumEntries = n;

    for (int i = 0; i < n; i ++)
    {
	npPal->palPalEntry[i].peRed = red[i];
	npPal->palPalEntry[i].peGreen = green[i];
	npPal->palPalEntry[i].peBlue = blue[i];
	npPal->palPalEntry[i].peFlags = 0;
    }
    ms_palette = CreatePalette((LPLOGPALETTE)npPal);
    LocalFree((HANDLE)npPal);
#endif  
    return TRUE;
}

int wxColourMap::GetPixel(const unsigned char red, const unsigned char green, const unsigned char blue)
{
  //  return ::GetNearestPaletteIndex(ms_palette, RGB(red, green, blue)); // MSW?
  return 0;
}

Bool wxColourMap::GetRGB(const int pixel, unsigned char *red, unsigned char *green, unsigned char *blue)
{
#if 0
  if (index < 0 || index > 255)
         return FALSE;

  PALETTEENTRY entry;
  if (::GetPaletteEntries(ms_palette, index, 1, &entry))
  {
         *red = entry.peRed;
         *green = entry.peGreen;
         *blue = entry.peBlue;
         return TRUE;
  } else
         return FALSE;
#endif

  return FALSE;
}

// Pens

IMPLEMENT_DYNAMIC_CLASS(wxPen, wxObject)

wxPen::wxPen(void)
{
  stipple = NULL ;
  style = wxSOLID;
  width = 1;
  join = wxJOIN_ROUND ;
  cap = wxCAP_ROUND ;
  nb_dash = 0 ;
  dash = NULL ;
  cpen = NULLHANDLE;
  my_old_cpen = NULLHANDLE ;
  old_width = -1 ;
  old_style = -1 ;
  old_join  = -1 ;
  old_cap  = -1 ;
  old_nb_dash  = -1 ;
  old_dash  = NULL ;
  old_color  = 0 ;
  old_stipple = NULL ;

  wxThePenList->AddPen(this);
}

wxPen::~wxPen()
{
    TRACE("wxPen::~wxPen() this=%#x", this);
    if (my_old_cpen) {
#if 0
	DeleteObject(my_old_cpen);
#endif    
    }
    my_old_cpen = NULLHANDLE ;

    if (cpen)
    {
#if 0
	DeleteObject(cpen);
#endif    
    }
    cpen = NULLHANDLE ;

    wxThePenList->RemovePen(this);
}

wxPen::wxPen(wxColour& col, int Width, int Style)
{
  colour = col;
  stipple = NULL;
  width = Width;
  style = Style;
  join = wxJOIN_ROUND ;
  cap = wxCAP_ROUND ;
  nb_dash = 0 ;
  dash = NULL ;
  cpen = NULLHANDLE ;
  my_old_cpen = NULLHANDLE ;
  old_width = -1 ;
  old_style = -1 ;
  old_join  = -1 ;
  old_cap  = -1 ;
  old_nb_dash  = -1 ;
  old_dash  = NULL ;
  old_color  = 0 ;
  old_stipple = NULL ;

#if 0
#ifndef WIN32
  // In Windows, only a pen of width = 1 can be dotted or dashed!
  if ((Style == wxDOT) || (Style == wxLONG_DASH) ||
      (Style == wxSHORT_DASH) || (Style == wxDOT_DASH) ||
      (Style == wxUSER_DASH))
    width = 1;
#else
/***
  DWORD vers = GetVersion() ;
  WORD  high = HIWORD(vers) ; // high bit=0 for NT, 1 for Win32s
  // Win32s doesn't support wide dashed pens

  if ((high&0x8000)!=0)
***/
  if (wxGetOsVersion()==wxWIN32S)
  {
    // In Windows, only a pen of width = 1 can be dotted or dashed!
    if ((Style == wxDOT) || (Style == wxLONG_DASH) ||
        (Style == wxSHORT_DASH) || (Style == wxDOT_DASH) ||
        (Style == wxUSER_DASH))
      width = 1;
  }
#endif 
#endif
  ChangePen() ;

  wxThePenList->AddPen(this);
}

wxPen::wxPen(const char *col, int Width, int Style)
{
  colour = col;
  stipple = NULL ;
  width = Width;
  style = Style;
  join = wxJOIN_ROUND ;
  cap = wxCAP_ROUND ;
  nb_dash = 0 ;
  dash = NULL ;
  cpen = NULLHANDLE ;
  my_old_cpen = NULLHANDLE ;
  old_width = -1 ;
  old_style = -1 ;
  old_join  = -1 ;
  old_cap  = -1 ;
  old_nb_dash  = -1 ;
  old_dash  = NULL ;
  old_color  = 0 ;
  old_stipple = NULL ;

#if 0
  // In Windows, only a pen of width = 1 can be dotted or dashed!
  if ((Style == wxDOT) || (Style == wxLONG_DASH) || (Style == wxSHORT_DASH) || (Style == wxDOT_DASH))
    width = 1;
#endif    
  ChangePen() ;

  wxThePenList->AddPen(this);
}

void wxPen::ChangePen(void)
{
    // Handled by wxPen::SelectPen() !!
    if (style==wxTRANSPARENT)
	return ;

    Bool must_change = FALSE ;

    COLOR os2_colour = 0 ;

    os2_colour = colour.pixel ;

    if (cpen==NULLHANDLE)
	must_change = TRUE ;
    else
	must_change = !(width==old_width     &&
			join==old_join       &&
			cap==old_cap         &&
			dash==old_dash       &&
			nb_dash==old_nb_dash &&
			style==old_style     &&
			stipple==old_stipple &&
			old_color==os2_colour) ;

    if (!must_change)
	return ;

    old_width = width ;
    old_join = join ;
    old_cap = cap ;
    old_dash = dash ;
    old_nb_dash = nb_dash ;
    old_style = style ;
    old_stipple = stipple ;
    old_color = os2_colour ;

/**
  No! if cpen is currently selected into some DC, this results into
  memory leakage with Win32s.
  Rather, defers Delete till next selection
  if (cpen)
  DeleteObject(cpen);
  cpen = NULL ;
  ***/
    if (cpen)
    {
	if (my_old_cpen)
	{
#if 0
	    DeleteObject(my_old_cpen) ; // and hope that's ok...
#endif      
	}
	my_old_cpen = cpen ;
	cpen = NULLHANDLE ;
    }

#if 0
    // Join style, Cap style, Pen Stippling only on Win32.
    // Currently no time to find equivalent on Win3.1, sorry
    // [if such equiv exist!!]
#ifdef WIN32
    if (join==wxJOIN_ROUND        &&
	cap==wxCAP_ROUND          &&
	style!=wxUSER_DASH        &&
	style!=wxSTIPPLE
	)
	cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);
    else
    {
	DWORD ms_style = PS_GEOMETRIC|wx2msPenStyle(style) ;

	LOGBRUSH logb ;

	switch(join)
	{
	case wxJOIN_BEVEL: ms_style |= PS_JOIN_BEVEL ; break ;
	case wxJOIN_MITER: ms_style |= PS_JOIN_MITER ; break ;
	default:
	case wxJOIN_ROUND: ms_style |= PS_JOIN_ROUND ; break ;
	}

	switch(cap)
	{
	case wxCAP_PROJECTING: ms_style |= PS_ENDCAP_SQUARE ; break ;
	case wxCAP_BUTT:       ms_style |= PS_ENDCAP_FLAT ;   break ;
	default:
	case wxCAP_ROUND:      ms_style |= PS_ENDCAP_ROUND ;  break ;
	}

	switch(style)
	{
	case wxSTIPPLE:
	    logb.lbStyle = BS_PATTERN ;
	    if (stipple)
		logb.lbHatch = (LONG)stipple->ms_bitmap ;
	    else
		logb.lbHatch = (LONG)0 ;
	    break ;
	case wxBDIAGONAL_HATCH:
	    logb.lbStyle = BS_HATCHED ;
	    logb.lbHatch = HS_BDIAGONAL ;
	    break ;
	case wxCROSSDIAG_HATCH:
	    logb.lbStyle = BS_HATCHED ;
	    logb.lbHatch = HS_DIAGCROSS ;
	    break ;
	case wxFDIAGONAL_HATCH:
	    logb.lbStyle = BS_HATCHED ;
	    logb.lbHatch = HS_FDIAGONAL ;
	    break ;
	case wxCROSS_HATCH:
	    logb.lbStyle = BS_HATCHED ;
	    logb.lbHatch = HS_CROSS ;
	    break ;
	case wxHORIZONTAL_HATCH:
	    logb.lbStyle = BS_HATCHED ;
	    logb.lbHatch = HS_HORIZONTAL ;
	    break ;
	case wxVERTICAL_HATCH:
	    logb.lbStyle = BS_HATCHED ;
	    logb.lbHatch = HS_VERTICAL ;
	    break ;
	default:
	    logb.lbStyle = BS_SOLID ;
	    break ;
	}
	logb.lbColor = ms_colour ;
	wxDash *real_dash ;
	if (style==wxUSER_DASH&&nb_dash&&dash)
	{
	    real_dash = new wxDash[nb_dash] ;
	    for (int i=0;i<nb_dash;i++)
		real_dash[i] = dash[i] * width ;
	}
	else
	    real_dash = NULL ;

/***
  DWORD vers = GetVersion() ;
  WORD  high = HIWORD(vers) ; // high bit=0 for NT, 1 for Win32s
  // Win32s doesn't have ExtCreatePen function...

  if ((high&0x8000)==0)
  ***/
	if (wxGetOsVersion()==wxWINDOWS_NT)
	    cpen = ExtCreatePen(ms_style,width,&logb,
				style==wxUSER_DASH?nb_dash:0,real_dash);
	else
	    cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);

	if (real_dash)
	    delete [] real_dash ;
    }
#else
    cpen = CreatePen(wx2msPenStyle(style), width, ms_colour);
#endif
#endif
#ifdef DEBUG_CREATE
    if (cpen==NULL) wxError("Cannot create pen","Internal error") ;
#endif

    return ;
}

void wxPen::SelectPen(HPS ps)
{
    LINEBUNDLE lb;

    // Set line attributes
    //lb.lColor  = colour ? colour : 0;
    lb.lColor  = colour;
    lb.lGeomWidth = MAKEFIXED((width==0) ? 1 : width, 0);
    
    // Set line pattern
    switch (style) {
    case wxDOT:
	lb.usType = LINETYPE_DOT;
	break;
    case wxSHORT_DASH:
	lb.usType = LINETYPE_SHORTDASH;
	break;
    case wxLONG_DASH:
	lb.usType = LINETYPE_LONGDASH;
	break;
    case wxTRANSPARENT:
	lb.usType = LINETYPE_INVISIBLE;
	break;
    case wxDOT_DASH:
	lb.usType = LINETYPE_DASHDOT;
	break;
    case wxSOLID:
    default:
	lb.usType = LINETYPE_SOLID;
	break;
    };

    // Set line end style
    switch(cap) {
    case wxCAP_PROJECTING: 
	lb.usEnd = LINEEND_SQUARE; break;
    case wxCAP_BUTT:
	lb.usEnd = LINEEND_FLAT; break;
    default:
    case wxCAP_ROUND:
	lb.usEnd = LINEEND_ROUND; break;
    };

    // Set line join style
    switch(join) {
    case wxJOIN_MITER:
	lb.usJoin = LINEJOIN_MITRE; break;
    case wxJOIN_BEVEL:
	lb.usJoin = LINEJOIN_BEVEL; break;
    default:
    case wxJOIN_ROUND:
	lb.usJoin = LINEJOIN_ROUND; break;
    };

    // Set RGB mode
    //  GpiCreateLogColorTable(ps, 0, LCOLF_RGB, 0, 0, NULL);
    GpiSetAttrs(ps, PRIM_LINE, LBB_COLOR|LBB_GEOM_WIDTH|LBB_TYPE|LBB_END|LBB_JOIN, 0, (PBUNDLE)&lb);
}

#if 0
int wx2os2PenStyle(int wx_style)
{
  int cstyle;
/***
#ifdef WIN32
  DWORD vers = GetVersion() ;
  WORD  high = HIWORD(vers) ; // high bit=0 for NT, 1 for Win32s
#endif
***/
  switch (wx_style)
  {
    case wxDOT:
      cstyle = PS_DOT;
      break;
    case wxSHORT_DASH:
    case wxLONG_DASH:
      cstyle = PS_DASH;
      break;
    case wxTRANSPARENT:
      cstyle = PS_NULL;
      break;
    case wxUSER_DASH:
      // User dash style not supported on Win3.1, sorry...
#ifdef WIN32
      // Win32s doesn't have PS_USERSTYLE
/***
      if ((high&0x8000)==0)
***/
      if (wxGetOsVersion()==wxWINDOWS_NT)
        cstyle = PS_USERSTYLE ;
      else
        cstyle = PS_DOT ; // We must make a choice... This is mine!
#else
      cstyle = PS_DASH ;
#endif
      break ;
    case wxSOLID:
    default:
      cstyle = PS_SOLID;
      break;
  }
  return cstyle;
}
#endif


// Brushes
IMPLEMENT_DYNAMIC_CLASS(wxBrush, wxObject)

wxBrush::wxBrush(void)
{
  cbrush = NULLHANDLE;
  my_old_cbrush = NULLHANDLE ;
  old_color = 0 ;
  old_style = -1 ;
  old_stipple = NULL ;

  wxTheBrushList->AddBrush(this);
}

wxBrush::~wxBrush()
{
    TRACE("wxBrush::~wxBrush() this=%#x", this);
    if (my_old_cbrush) {
#if 0
	DeleteObject(my_old_cbrush);
#endif
    }
    my_old_cbrush = NULLHANDLE ;

    if (cbrush)
    {
#if 0
	DeleteObject(cbrush);
#endif
    }
    cbrush = NULLHANDLE ;

    wxTheBrushList->RemoveBrush(this);
}

wxBrush::wxBrush(wxColour& col, int Style)
{
  colour = col;
  style = Style;
  stipple = NULL ;
  cbrush = NULLHANDLE;
  my_old_cbrush = NULLHANDLE;
  old_color = 0 ;
  old_style = -1 ;
  old_stipple = NULL ;

  ChangeBrush();
  
  wxTheBrushList->AddBrush(this);
}

void wxBrush::ChangeBrush(void) 
{
    // Handled by wxBrush::SelectBrush() !!
    if (style==wxTRANSPARENT)
	return ;

    Bool must_change = FALSE ;

    COLOR os2_colour = colour ;

    if (cbrush==NULLHANDLE)
	must_change = TRUE ;
    else
	must_change = !(style==old_style     &&
			stipple==old_stipple &&
			old_color==os2_colour) ;

    if (!must_change)
	return ;

/***
  No! (see ChangePen)
  if (cbrush) DeleteObject(cbrush);
  cbrush = NULL ;
  ***/

    if (cbrush) {
	if (my_old_cbrush) {
#if 0
	    DeleteObject(my_old_cbrush) ;
#endif
	}
	my_old_cbrush = cbrush ;
	cbrush = NULLHANDLE ;
    }

#if 0
    switch (style)
    {
/****
  // Don't reset cbrush, wxTRANSPARENT is handled by wxBrush::SelectBrush()
  // this could save (many) time if frequently switching from
  // wxSOLID to wxTRANSPARENT, because Create... is not always called!!
  //
  case wxTRANSPARENT:
  cbrush = NULL;  // Must always select a suitable background brush
  // - could choose white always for a quick solution
  break;
  ***/
    case wxBDIAGONAL_HATCH:
	cbrush = CreateHatchBrush(HS_BDIAGONAL,ms_colour) ;
	break ;
    case wxCROSSDIAG_HATCH:
	cbrush = CreateHatchBrush(HS_DIAGCROSS,ms_colour) ;
	break ;
    case wxFDIAGONAL_HATCH:
	cbrush = CreateHatchBrush(HS_FDIAGONAL,ms_colour) ;
	break ;
    case wxCROSS_HATCH:
	cbrush = CreateHatchBrush(HS_CROSS,ms_colour) ;
	break ;
    case wxHORIZONTAL_HATCH:
	cbrush = CreateHatchBrush(HS_HORIZONTAL,ms_colour) ;
	break ;
    case wxVERTICAL_HATCH:
	cbrush = CreateHatchBrush(HS_VERTICAL,ms_colour) ;
	break ;
    case wxSTIPPLE:
	if (stipple)
	    cbrush = CreatePatternBrush(stipple->ms_bitmap) ;
	else
	    cbrush = CreateSolidBrush(ms_colour) ;
	break ;
    case wxSOLID:
    default:
	cbrush = CreateSolidBrush(ms_colour) ;
	break;
    }
#endif
    old_style = style ;
    old_stipple = stipple ;
    old_color = os2_colour ;
#ifdef DEBUG_CREATE
    if (cbrush==NULL) wxError("Cannot create brush","Internal error") ;
#endif

}

void wxBrush::SelectBrush(HPS ps)
{
    AREABUNDLE ab;

    //ab.lColor = colour ? colour : 0;
    ab.lColor = colour;
    switch(style) {
    case wxHORIZONTAL_HATCH:
	ab.usSymbol = PATSYM_HORIZ; break;
    case wxVERTICAL_HATCH:
	ab.usSymbol = PATSYM_VERT; break;
    default:
    case wxSOLID:
	ab.usSymbol = PATSYM_SOLID; break;
    };
    GpiSetAttrs(ps, PRIM_AREA, ABB_COLOR | ABB_SYMBOL, 0, (PBUNDLE)&ab);
}

wxBrush::wxBrush(const char *col, int Style)
{
  colour = col;
  style = Style;
  stipple = NULL ;
  cbrush = NULLHANDLE;
  my_old_cbrush = NULLHANDLE;
  old_color = 0 ;
  old_style = -1 ;
  old_stipple = NULL ;

  wxTheBrushList->AddBrush(this);
}

// Icons
IMPLEMENT_DYNAMIC_CLASS(wxIcon, wxBitmap)

wxIcon::wxIcon(void)
{
    __type = wxTYPE_ICON;
    width = 0;
    height = 0;
    os2_icon = NULLHANDLE ;
    ok = FALSE;
//  wxTheIconList->Append(this);
}

wxIcon::wxIcon(char bits[], int width, int height)
{
    __type = wxTYPE_ICON;
    width = 0;
    height = 0;
    os2_icon = NULLHANDLE ;
    ok = FALSE;
//  wxTheIconList->Append(this);
}

wxIcon::wxIcon(const char *icon_file, long flags)
{
    __type = wxTYPE_ICON;
    width = 0 ;
    height = 0 ;
    LoadFile((char *)icon_file, flags);
//  wxTheIconList->Append(this);
}

wxIcon::wxIcon(ULONG idPointer)
{
    TRACE("wxIcon::wxIcon(idPointer=%d)", (int)idPointer);
    __type = wxTYPE_ICON;
    os2_icon = WinLoadPointer(HWND_DESKTOP, NULLHANDLE, idPointer);
    POINTERINFO pointerinfo;
    WinQueryPointerInfo(os2_icon, &pointerinfo);
    BITMAPINFOHEADER bmpData;
    GpiQueryBitmapParameters(pointerinfo.hbmPointer, &bmpData);
    width = bmpData.cx;
    height = bmpData.cy;
//  wxTheIconList->Append(this);
}

wxIcon::~wxIcon(void)
{
    TRACE("wxIcon::~wxIcon() this=%#x wxBitmap=%#x", this, (wxBitmap *)this);
    if (os2_icon)
	WinDestroyPointer(os2_icon);

//  wxTheIconList->DeleteObject(this);
}

Bool wxIcon::LoadFile(char *icon_file, long flags)
{
    TRACE("wxIcon::LoadFile(icon_file='%s', flags=%#x)", icon_file, (int)flags);

    __type = wxTYPE_ICON;
    ok = FALSE;
    width = 0;
    height = 0;
    depth = 0;

    if (flags & wxBITMAP_TYPE_ICO_RESOURCE)
	wxFatalError("wxIcon: Cannot load icons by name from resource file on OS/2");

#if USE_RESOURCE_LOADING_IN_OS2
    else if (flags & wxBITMAP_TYPE_ICO) {
	os2_icon = ReadIconFile(icon_file, wxhAb, NULL, NULL);
	ok = (ms_icon != 0);
	return ok;
    }
#endif
    wxError("Unimplemented icon type");
    return FALSE;
}

// Cursors
IMPLEMENT_DYNAMIC_CLASS(wxCursor, wxBitmap)

wxCursor::wxCursor(void)
{
    __type = wxTYPE_CURSOR;
    width = 32; height = 32;
    os2_cursor = NULLHANDLE ;
    destroyCursor = FALSE;
//  wxTheCursorList->Append(this) ;
}

wxCursor::wxCursor(char WXUNUSED(bits)[], int width, int height)
{
    __type = wxTYPE_CURSOR;
    os2_cursor = NULLHANDLE ;
    destroyCursor = FALSE;
//  wxTheCursorList->Append(this) ;
}

wxCursor::wxCursor(const char *cursor_file, long flags, int hotSpotX, int hotSpotY)
{
    __type = wxTYPE_CURSOR;
    destroyCursor = FALSE;
    os2_cursor = NULLHANDLE;
    ok = FALSE;
    if (flags & wxBITMAP_TYPE_CUR_RESOURCE) {
	wxFatalError("wxCursor: Cannot load cursors by name from resource file on OS/2");
    } else if (flags & wxBITMAP_TYPE_CUR) {
#if USE_RESOURCE_LOADING_IN_OS2
	os2_cursor = ReadCursorFile((char *)cursor_file, wxhInstance, &width, &height);
#endif
    } else if (flags & wxBITMAP_TYPE_ICO) {
#if USE_RESOURCE_LOADING_IN_OS2
	ms_cursor = IconToCursor((char *)cursor_file, wxhInstance, hotSpotX, hotSpotY, &width, &height);
#endif
    } else if (flags & wxBITMAP_TYPE_BMP) {
#if USE_RESOURCE_LOADING_IN_OS2
	HBITMAP hBitmap = 0;
	HPALETTE hPalette = 0;
	Bool success = ReadDIB((char *)cursor_file, &hBitmap, &hPalette);
	if (!success)
	    return;
	if (hPalette)
	    DeleteObject(hPalette);
	POINT pnt;
	pnt.x = hotSpotX;
	pnt.y = hotSpotY;
	ms_cursor = MakeCursorFromBitmap(wxhInstance, hBitmap, &pnt);
	DeleteObject(hBitmap);
	if (ms_cursor)
	    ok = TRUE;
#endif
    }
//  wxTheCursorList->Append(this) ;
}

// Cursors by stock number
wxCursor::wxCursor(int cursor_type)
{
    __type = wxTYPE_CURSOR;
    switch (cursor_type) {
    case wxCURSOR_WAIT:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_WAIT, FALSE);
	break;
    case wxCURSOR_IBEAM:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_TEXT, FALSE);
	break;
    case wxCURSOR_CROSS:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_CROSS);
	break;
    case wxCURSOR_SIZENWSE:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_SIZENWSE, FALSE);
	break;
    case wxCURSOR_SIZENESW:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_SIZENESW, FALSE);
	break;
    case wxCURSOR_SIZEWE:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_SIZEWE, FALSE);
	break;
    case wxCURSOR_SIZENS:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_SIZENS, FALSE);
	break;
    case wxCURSOR_CHAR:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_ARROW, FALSE);
	break;
    case wxCURSOR_HAND:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_HAND);
	break;
    case wxCURSOR_BULLSEYE:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_BULLSEYE);
	break;
    case wxCURSOR_PENCIL:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_PENCIL);
	break;
    case wxCURSOR_MAGNIFIER:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_MAGNIFIER);
	break;
    case wxCURSOR_NO_ENTRY:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_ILLEGAL, FALSE);
	break;
    case wxCURSOR_LEFT_BUTTON:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_ARROW, FALSE);
	break;
    case wxCURSOR_RIGHT_BUTTON:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_ARROW, FALSE);
	break;
    case wxCURSOR_MIDDLE_BUTTON:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_ARROW, FALSE);
	break;
    case wxCURSOR_SIZING:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_MOVE, FALSE);
	break;
    case wxCURSOR_WATCH:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_WATCH);
	break;
    case wxCURSOR_SPRAYCAN:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_ROLLER);
	break;
    case wxCURSOR_PAINT_BRUSH:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_PBRUSH);
	break;
    case wxCURSOR_POINT_LEFT:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_PLEFT);
	break;
    case wxCURSOR_POINT_RIGHT:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_PRIGHT);
	break;
    case wxCURSOR_QUESTION_ARROW:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_QARROW);
	break;
#if 0
    case wxCURSOR_BLANK:
	os2_cursor = WinLoadPointer(HWND_DESKTOP, 0, _wxCURSOR_BLANK);
	break;
#endif
    default:
    case wxCURSOR_ARROW:
	os2_cursor = WinQuerySysPointer(HWND_DESKTOP, SPTR_ARROW, FALSE);
	break;
    }
//  wxTheCursorList->Append(this) ;
}

wxCursor::~wxCursor(void)
{
    TRACE("wxCursor::~wxCursor() this=%#x wxBitmap=%#x", this, (wxBitmap *)this);
//  wxTheCursorList->DeleteObject(this) ;
}

// Global cursor setting
void wxSetCursor(wxCursor *cursor)
{
    TRACE("wxSetCursor(cursor=%#x, cursor->os2_cursor=%#x", (int)cursor, (int)cursor->os2_cursor);
//
// XXX
// Currently has no effect in OS/2
//
//    if (cursor && cursor->os2_cursor)
//	WinSetPointer(HWND_DESKTOP, cursor->os2_cursor);

    wxFlushEvents();
}

// Misc. functions

// Return TRUE if we have a colour display
Bool wxColourDisplay(void)
{
    LONG clrs;
    HPS ps  = WinGetPS(HWND_DESKTOP);
    HDC hdc = GpiQueryDevice(ps);
    DevQueryCaps(hdc, CAPS_COLORS, 1L, &clrs);
    WinReleasePS(ps);
    return (clrs > 2) ? TRUE : FALSE;
}

// Returns depth of screen
int wxDisplayDepth(void)
{
    LONG depth;
    HPS ps  = WinGetPS(HWND_DESKTOP);
    HDC hdc = GpiQueryDevice(ps);
    DevQueryCaps(hdc, CAPS_COLOR_PLANES, 1L, &depth);
    WinReleasePS(ps);
    return (int)depth;
}

// Get size of display
void wxDisplaySize(int *width, int *height)
{
    LONG hres, vres;
    HPS ps  = WinGetPS(HWND_DESKTOP);
    HDC hdc = GpiQueryDevice(ps);
    DevQueryCaps(hdc, CAPS_WIDTH, 1L, &hres);
    DevQueryCaps(hdc, CAPS_HEIGHT, 1L, &vres);
    WinReleasePS(ps);
    *width  = hres;
    *height = vres;
}

IMPLEMENT_DYNAMIC_CLASS(wxBitmap, wxObject)

wxBitmap::wxBitmap(void)
{
    __type = wxTYPE_BITMAP;
    ok = FALSE;
    width = 0;
    height = 0;
    depth = 0;
    os2_bitmap = NULLHANDLE ;
    selectedInto = NULL;
    numColors = 0;
    bitmapColourMap = NULL;
    wxTheBitmapList->Append(this);
}

wxBitmap::wxBitmap(char bits[], int the_width, int the_height, int no_bits)
{
    TRACE("wxBitmap::wxBitmap(..., the_width=%d, the_height=%d, no_bits=%d)", the_width, the_height, no_bits);
    __type = wxTYPE_BITMAP;
    width = the_width;
    height = the_height;
    depth = no_bits;
    numColors = 0;
    bitmapColourMap = NULL;
#if 0
    HPS ps = WinGetPS(HWND_DESKTOP);
#else
    SIZEL sizl = {0, 0};
#if 0
    DEVOPENSTRUC dop = {0L, (PSZ)"DISPLAY", NULL, 0L, 0L, 0L, 0L, 0L, 0L };
    HDC hdc = DevOpenDC(wxhAb, OD_MEMORY, (PSZ)"*", 5L, (PDEVOPENDATA)&dop, NULLHANDLE);
#else
    HDC hdc = DevOpenDC(wxhAb, OD_MEMORY, (PSZ)"*", 0L, NULL, NULLHANDLE);
#endif
    HPS ps = GpiCreatePS(wxhAb, hdc, &sizl, PU_PELS | GPIA_ASSOC);
#endif

    BITMAPINFOHEADER2 bmih;  /* bit map info structure               */
                                                         
    memset(&bmih, 0, sizeof(BITMAPINFOHEADER2));
    bmih.cbFix     = sizeof(BITMAPINFOHEADER2);
    bmih.cy        = the_height;
    bmih.cx        = the_width;
    bmih.cPlanes   = 1;
    bmih.cBitCount = no_bits;

    os2_bitmap = GpiCreateBitmap(ps, &bmih, 0, (PBYTE)bits, NULL);
    TRACE("wxBitmap::wxBitmap: os2_bitmap=%#x\n", os2_bitmap);

#if 0
    WinReleasePS(ps);
#else
    GpiDestroyPS(ps);
    DevCloseDC(hdc);
#endif
                                                      
    if (os2_bitmap)
	ok = TRUE;
    else
	ok = FALSE;

    selectedInto = NULL;
    wxTheBitmapList->Append(this);
}

wxBitmap::wxBitmap(int w, int h, int d)
{
    __type = wxTYPE_BITMAP;
    ok = FALSE;
    width = w;
    height = h;
    depth = d;
    numColors = 0;
    selectedInto = NULL;
    bitmapColourMap = NULL;

    (void)Create(w, h, d);
    wxTheBitmapList->Append(this);
}

wxBitmap::wxBitmap(char *bitmap_file, long flags)
{
    __type = wxTYPE_BITMAP;
    ok = FALSE;
    width = 0;
    height = 0;
    depth = 0;
    selectedInto = NULL;
    numColors = 0;
    bitmapColourMap = NULL;

    LoadFile(bitmap_file, (int)flags);
    wxTheBitmapList->Append(this);
}

wxBitmap::wxBitmap(ULONG resourceId)
{
    TRACE("wxBitmap::wxBitmap(resourceId=%d)", (int)resourceId);
    __type = wxTYPE_BITMAP;
    ok = FALSE;
    depth = 0;
    selectedInto = NULL;
    numColors = 0;
    bitmapColourMap = NULL;
    wxTheBitmapList->Append(this);

    BITMAPINFOHEADER bmInfo;
    HPS screenPs = WinGetPS(HWND_DESKTOP);

    os2_bitmap = GpiLoadBitmap(screenPs, NULLHANDLE, resourceId, 0L, 0L);
TRACE("wxBitmap::wxBitmap: os2_bitmap=%#x", os2_bitmap);
    GpiQueryBitmapParameters(os2_bitmap, &bmInfo);
    width  = bmInfo.cx;
    height = bmInfo.cy;
    //depth  = bmInfo.depth;
    WinReleasePS(screenPs);

    wxTheBitmapList->Append(this);
}      


#if USE_XPM_IN_OS2
// Create from data
wxBitmap::wxBitmap(char **data, wxItem *anItem)
{
  __type = wxTYPE_BITMAP;
  selectedInto = NULL;
  bitmapColourMap = NULL;

  XImage *ximage;
  int     ErrorStatus;
  XpmAttributes xpmAttr;
  HDC     dc;

  ok = FALSE;
  numColors = 0;

  dc = CreateCompatibleDC(NULL);        /* memory DC */

  if (dc)
  {
    xpmAttr.valuemask = XpmReturnInfos;        /* get infos back */
    ErrorStatus = XpmCreateImageFromData(&dc, data,
         &ximage, (XImage **) NULL, &xpmAttr);

    if (ErrorStatus == XpmSuccess)
    {
      /* ximage is malloced and contains bitmap and attributes */
      ms_bitmap = ximage->bitmap;

      BITMAP  bm;
      GetObject(ms_bitmap, sizeof(bm), (LPSTR) & bm);

      width = (bm.bmWidth);
      height = (bm.bmHeight);
      depth = (bm.bmPlanes * bm.bmBitsPixel);
      numColors = xpmAttr.npixels;
      XpmFreeAttributes(&xpmAttr);

      XImageFree(ximage);        // releases the malloc, but does not detroy
                        // the bitmap
      ok = TRUE;

    } else
    {
      ok = False;
//  XpmDebugError(ErrorStatus, NULL);
    }
    DeleteDC(dc);
  }
}
#endif

Bool wxBitmap::Create(int w, int h, int d)
{
    TRACE("wxBitmap::Create(w=%d, h=%d, d=%d)", w, h, d);
    width = w;
    height = h;
#if 0
    HPS ps = WinGetPS(HWND_DESKTOP);
#else
    SIZEL sizl = {0, 0};
#if 0
    DEVOPENSTRUC dop = {0L, (PSZ)"DISPLAY", NULL, 0L, 0L, 0L, 0L, 0L, 0L };
    HDC hdc = DevOpenDC(wxhAb, OD_MEMORY, (PSZ)"*", 5L, (PDEVOPENDATA)&dop, NULLHANDLE);
#else
    HDC hdc = DevOpenDC(wxhAb, OD_MEMORY, (PSZ)"*", 0L, NULL, NULLHANDLE);
#endif
    HPS ps = GpiCreatePS(wxhAb, hdc, &sizl, PU_PELS | GPIA_ASSOC);
#endif

    LONG planes;
    if (d <= 0) {
	LONG bitcount;
#if 0
	HDC hdc = GpiQueryDevice(ps);
#endif
	DevQueryCaps(hdc, CAPS_COLOR_PLANES, 1L, &planes);
	DevQueryCaps(hdc, CAPS_COLOR_BITCOUNT, 1L, &bitcount);
TRACE("CAPS_COLOR_PLANES=%d, CAPS_COLOR_BITCOUNT=%d", planes, bitcount);
	depth = bitcount;
    } else {
	planes = 1;
	depth = d;
    }
    BITMAPINFOHEADER2 bmih;		/* bit map info structure               */
                                                         
    memset(&bmih,0, sizeof(BITMAPINFOHEADER2));
    bmih.cbFix     = sizeof(BITMAPINFOHEADER2);
    bmih.cy        = height;
    bmih.cx        = width;
    bmih.cPlanes   = planes;
    bmih.cBitCount = depth;
TRACE("wxBitmap::Create: Creating bitmap width=%d, height=%d, planes=%d, depth=%d", width, height, planes, depth);
    os2_bitmap = GpiCreateBitmap(ps, &bmih, 0, NULL, NULL);
    TRACE("wxBitmap::Create: os2_bitmap=%#x", os2_bitmap);

    assert(os2_bitmap != NULLHANDLE);

#if 0
    WinReleasePS(ps);
#else
    GpiDestroyPS(ps);
    DevCloseDC(hdc);
#endif

    if (os2_bitmap)
	ok = TRUE;
    else
	ok = FALSE;
    return ok;
}

Bool wxBitmap::LoadFile(char *bitmap_file, long flags)  
{
    ok = FALSE;
    width = 0;
    height = 0;
    depth = 0;
    selectedInto = NULL;
  
#if 0
    if (flags & wxBITMAP_TYPE_BMP_RESOURCE) {
	ms_bitmap = LoadBitmap(wxhAb, bitmap_file);
	if (ms_bitmap)
	{
	    ok = TRUE;
	    BITMAP bm;
	    GetObject(ms_bitmap, sizeof(BITMAP), (LPSTR) &bm);
	    width = bm.bmWidth;
	    height = bm.bmHeight;
	    depth = bm.bmPlanes;
	    return TRUE;
	}
    }
#endif
#if USE_XPM_IN_OS2
    else if (flags & wxBITMAP_TYPE_XPM) {
	XImage *ximage;
	XpmAttributes xpmAttr;
	HDC     dc;

	ok = False;
	dc = CreateCompatibleDC(NULL);
	if (dc)
	{
	    xpmAttr.valuemask = XpmReturnPixels;
	    int errorStatus = XpmReadFileToImage(&dc, bitmap_file, &ximage, (XImage **) NULL, &xpmAttr);
	    DeleteDC(dc);
	    if (errorStatus == XpmSuccess)
	    {
		ms_bitmap = ximage->bitmap;

		BITMAP  bm;
		GetObject(ms_bitmap, sizeof(bm), (LPSTR) & bm);

		width = (bm.bmWidth);
		height = (bm.bmHeight);
		depth = (bm.bmPlanes * bm.bmBitsPixel);
		numColors = xpmAttr.npixels;
		XpmFreeAttributes(&xpmAttr);
		XImageFree(ximage);
        
		ok = TRUE;
		return TRUE;
	    }
	    else
	    {
		ok = FALSE;
		return FALSE;
	    }
	}
    }
#endif
#if USE_IMAGE_LOADING_IN_OS2
    else if ((flags & wxBITMAP_TYPE_BMP) || (flags & wxBITMAP_TYPE_ANY))
    {
	wxColourMap *cmap = NULL;
	Bool success = FALSE;
	if (flags & wxBITMAP_DISCARD_COLOURMAP)
	    success = wxLoadIntoBitmap(bitmap_file, this);
	else
	    success = wxLoadIntoBitmap(bitmap_file, this, &cmap);
	if (!success && cmap)
	{
	    delete cmap;
	    cmap = NULL;
	}
	if (cmap)
	    bitmapColourMap = cmap;
	return success;
    }
#endif
    return FALSE;
}

wxBitmap::~wxBitmap(void)
{
    TRACE("wxBitmap::~wxBitmap() this=%#x", this);
    if (selectedInto) {
	char buf[200];
	sprintf(buf, "Bitmap %X was deleted without selecting out of wxMemoryDC %X.", (int)this, (int)selectedInto);
	wxFatalError(buf);
    }

    if (os2_bitmap)
	GpiDeleteBitmap(os2_bitmap);
    os2_bitmap = NULLHANDLE;

    if (bitmapColourMap)
	delete bitmapColourMap;

    wxTheBitmapList->DeleteObject(this);
}

Bool wxBitmap::SaveFile(char *filename, int typ, wxColourMap *cmap)
{
  switch (typ)
  {
#if USE_IMAGE_LOADING_IN_OS2
    case wxBITMAP_TYPE_BMP:
    {
      wxColourMap *actualCmap = cmap;
      if (!actualCmap)
        actualCmap = bitmapColourMap;
      return wxSaveBitmap(filename, this, actualCmap);
      break;
    }
#endif
#if USE_XPM_IN_OS2
    case wxBITMAP_TYPE_XPM:
    {
      HDC     dc = NULL;

      Visual *visual = NULL;
      XImage  ximage;

      dc = CreateCompatibleDC(NULL);
      if (dc)
      {
        if (SelectObject(dc, ms_bitmap))
        { /* for following SetPixel */
          /* fill the XImage struct 'by hand' */
          ximage.width = width; ximage.height = height;
          ximage.depth = depth; ximage.bitmap = ms_bitmap;
          int errorStatus = XpmWriteFileFromImage(&dc, filename,
                                                  &ximage, (XImage *) NULL, (XpmAttributes *) NULL);

          if (dc)
            DeleteDC(dc);

          if (errorStatus == XpmSuccess)
            return TRUE;                /* no error */
          else
            return FALSE;
        } else return FALSE;
      } else return FALSE;
      break;
    }
#endif
    default:
      break;
  }
  return FALSE;
}


// Return the system colour for a particular GUI element
wxColour wxGetSysColour(int sysColour)
{
    // OS/2 colour ids the same as wxWindows ones, so no translation.
    ULONG rgbColour = WinQuerySysColor(HWND_DESKTOP, sysColour, 0);
    TRACE("wxGetSysColour(%d) returns %#06x", rgbColour);
    return wxColour(rgbColour);
}

/*
 * Get the color assigned with the indicated presentation
 * parameter, or -1 for not assigned.
 */
COLOR wxGetForegroundColour(HWND hwnd)
{
    ULONG clr;
    if (WinQueryPresParam(hwnd, PP_FOREGROUNDCOLOR, 0, NULL, sizeof(clr), &clr, QPF_PURERGBCOLOR))
	return clr;
    else
	return -1;
}

COLOR wxGetBackgroundColour(HWND hwnd)
{
    ULONG clr;
    if (WinQueryPresParam(hwnd, PP_BACKGROUNDCOLOR, 0, NULL, sizeof(clr), &clr, QPF_PURERGBCOLOR))
	return clr;
    else
	return -1;
}

void wxSetBackgroundColour(HWND hwnd, COLOR clr)
{
    TRACE("wxSetBackgroundColour(hwnd=%#x, clr=%#06x)", hwnd, clr);
    WinSetPresParam(hwnd, PP_BACKGROUNDCOLOR, sizeof(clr), &clr);
}

void wxSetForegroundColour(HWND hwnd, COLOR clr)
{
    TRACE("wxSetForegroundColour(hwnd=%#x, clr=%#06x)", hwnd, clr);
    WinSetPresParam(hwnd, PP_FOREGROUNDCOLOR, sizeof(clr), &clr);
}

void wxSetWinFont(HWND hwnd, const char *fontname)
{
    TRACE("wxSetWinFont(hwnd=%#x, fontname='%s')", hwnd, fontname);
    if (fontname)
	WinSetPresParam(hwnd, PP_FONTNAMESIZE, strlen(fontname) + 1, (PVOID)fontname);
}

void wxSetWinFont(HWND hwnd, wxFont *font)
{
    wxSetWinFont(hwnd, font->GetNameSize());
}

