/*
 * File:	wx_slidr.cc
 * Purpose:	Slider implementation
 * Author:	Julian Smart
 * Created:	1993
 * Updated:	April 1995
 * Copyright:	(c) 1995, AIAI, University of Edinburgh
 */

/* static const char sccsid[] = "%W% %G%"; */

#include "wx_panel.h"
#include "wx_slidr.h"
#include "wx_itemp.h"

wxList wxScrollBarList;

// Slider
IMPLEMENT_DYNAMIC_CLASS(wxSlider, wxItem)

wxSlider::wxSlider(void)
{
    hwndEditValue = NULLHANDLE;
    hwndStaticMin = NULLHANDLE;
    page_size = 0;
    s_min = s_max = 0;
    hwndStaticMax = 0;
}

wxSlider::wxSlider(wxPanel *panel, wxFunction func, const char *label, int value,
		   int min_value, int max_value, int width, int x, int y,
		   long style, const char *name)
    : wxbSlider(panel, func, label, value, min_value, max_value, width, x, y, style, name)
{
    Create(panel, func, label, value, min_value, max_value, width, x, y, style, name);
}

Bool wxSlider::Create(wxPanel *panel, wxFunction func, const char *label, int value,
           int min_value, int max_value, int width, int x, int y, long style, const char *name)
{
    TRACE("wxSlider::Create(panel=%#x, func=%#x, label='%s', value=%d, min_value=%d, max+value=%d, w=%d, x=%d, y=%d, style=%#x, name='%s')", (int)panel, (int)func, label, value, min_value, max_value, width, x, y, (int)style, name);
    SetName(name);
    if (panel) panel->AddChild(this);
    buttonFont = panel->buttonFont ;
    labelFont = panel->labelFont ;
    backColour = panel->backColour ;
    labelColour = panel->labelColour ;
    buttonColour = panel->buttonColour ;

    windowStyle = style;

    labelPosition = panel->label_position;
    panel->GetValidPosition(&x, &y);

    int cx;
    int cy;
    wxGetCharSize(panel->GetHWND(), &cx, &cy);

    const char *cf;
    // If label exists, create a static control for it.
    if (label)
	CreateLabel(panel->GetHWND(), label);
    
    hwndEditValue = WinCreateWindow(panel->GetHWND(), WC_ENTRYFIELD, NULL, ES_AUTOSCROLL | ES_LEFT |
				 WS_TABSTOP, 0, 0, 0, 0, panel->GetHWND(),
				 HWND_TOP, NewControlId(), NULL, NULL);

    // Now create min static control
    sprintf(wxBuffer, "%d", min_value);
    hwndStaticMin =  WinCreateWindow(panel->GetHWND(), WC_STATIC, (PSZ)wxBuffer, SS_TEXT,
				  0, 0, 0, 0, panel->GetHWND(),
				  HWND_TOP, NewControlId(), NULL, NULL);

    wxSetBackgroundColour(hwndStaticMin, backColour);
    wxSetForegroundColour(hwndStaticMin, labelColour);

    // Now create slider
    ULONG sstyle = 0;
    if (windowStyle & wxVERTICAL)
	sstyle = SBS_VERT;
    else
	sstyle = SBS_HORZ;

    windows_id = (int)NewControlId();
    hwndItem = WinCreateWindow(panel->GetHWND(), WC_SCROLLBAR, (PSZ)wxBuffer,
				 sstyle, 0, 0, 0, 0, panel->GetHWND(),
				 HWND_TOP, windows_id, NULL, NULL); 
    TRACE("wxSlider::Create: Window created, parent=%#x, handle=%#x", panel->GetHWND(), hwndItem);

    // Subclass again for purposes of dialog editing mode
    SubclassControl(hwndItem);

    wxScrollBarList.Append(windows_id, this);

    page_size = (int)((max_value-min_value)/10);
    s_max = max_value;
    s_min = min_value;
    WinSendMsg(hwndItem, SBM_SETSCROLLBAR, (MPARAM)value, MPFROM2SHORT(0, max_value - min_value));

    // Finally, create max value static item
    sprintf(wxBuffer, "%d", max_value);
    hwndStaticMax =  WinCreateWindow(panel->GetHWND(), WC_STATIC, (PSZ)wxBuffer, SS_TEXT,
				  0, 0, 0, 0, panel->GetHWND(),
				  HWND_TOP, NewControlId(), NULL, NULL);

    wxSetBackgroundColour(hwndStaticMax, backColour);
    wxSetForegroundColour(hwndStaticMax, labelColour);

    if (buttonFont) {
	wxSetWinFont(hwndStaticMin, buttonFont);
	wxSetWinFont(hwndStaticMax, buttonFont);
	wxSetWinFont(hwndEditValue, buttonFont);
    }

    SetSize(x, y, width, -1);
    Show();
    SetValue(value);

    panel->AdvanceCursor(this);
    Callback(func);

    return TRUE;
}

// Called from wx_win.cc: wxWnd::OnHScroll, wxWnd::OnVScroll
void wxSliderEvent(int usidentifier, int uscmd, int sslider)
{
    TRACE("wxSliderEvent(usidentifier=%d, uscmd=%#x, sslider=%d)", usidentifier, uscmd, sslider);
    wxNode *node = (wxNode *)wxScrollBarList.Find((long)usidentifier);
    if (!node)
	return;

    wxSlider *slider = (wxSlider *)node->Data();

    int position = slider->GetValue();
    TRACE("\tposition = %d", position);
    int nScrollInc;

    switch (uscmd) {
    case SB_LINEUP:
	nScrollInc = -1;
	break;

    case SB_LINEDOWN:
	nScrollInc = 1;
	break;

    case SB_PAGEUP:
	nScrollInc = -slider->page_size;
	break;

    case SB_PAGEDOWN:
	nScrollInc = slider->page_size;;
	break;

    case SB_SLIDERTRACK:
	nScrollInc = (slider->s_min + sslider) - position;
	break;

    default:
	nScrollInc = 0;
    }

    TRACE("\tnScrollInc = %d", nScrollInc);
    if (nScrollInc != 0) {
	int new_pos = position + nScrollInc;
	TRACE("new_pos=%d", new_pos);
	if (new_pos >= slider->s_min && new_pos < slider->s_max) {
	    slider->SetValue(new_pos);
	    wxCommandEvent event(wxEVENT_TYPE_SLIDER_COMMAND);
	    event.commandInt = new_pos;
	    event.eventObject = slider;
	    slider->ProcessCommand(event);
	}
    }
}

wxSlider::~wxSlider(void)
{
    TRACE("wxSlider::~wxSlider() this=%#x wxWindow=%#x", this, (wxWindow *)this);
    if (hwndStaticMin)
	WinDestroyWindow(hwndStaticMin);
    if (hwndStaticMax)
	WinDestroyWindow(hwndStaticMax);
    if (hwndEditValue)
	WinDestroyWindow(hwndEditValue);
    wxScrollBarList.DeleteObject(this);
}

int wxSlider::GetValue(void)
{
    return (int)WinSendMsg(hwndItem, SBM_QUERYPOS, (MPARAM)NULL, (MPARAM)NULL) + s_min;
}

void wxSlider::SetValue(int value)
{
    TRACE("wxSlider::SetValue(%d) this=%#x", value, (int)this);
    if (value < s_min)
	value = s_min;
    else if (value > s_max)
	value = s_max;
    WinSendMsg(hwndItem, SBM_SETPOS, (MPARAM)value - s_min, (MPARAM)0);
    sprintf(wxBuffer, "%d", value);
    WinSetWindowText(hwndEditValue, (PSZ)wxBuffer);
}

void wxSlider::GetSize(int *width, int *height)
{
    wxRECTL rect = {-1, -1, -1, -1};

    wxFindMaxSize(hwndItem, &rect);
    if (HaveLabel())
	wxFindMaxSize(GetLabelHWND(), &rect);
    wxFindMaxSize(hwndStaticMin, &rect);
    wxFindMaxSize(hwndStaticMax, &rect);
    wxFindMaxSize(hwndEditValue, &rect);

    *width  = rcWidth(rect);
    *height = rcHeight(rect);
    TRACE("wxSlider::GetSize() returns width=%d, height=%d", *width, *height);
}

void wxSlider::GetPosition(int *x, int *y)
{
    wxWindow *parent = GetParent();
    wxRECTL rect = {-1, -1, -1, -1};

    wxFindMaxSize(hwndItem, &rect);

    if (HaveLabel())
	wxFindMaxSize(GetLabelHWND(), &rect);
    wxFindMaxSize(hwndStaticMin, &rect);
    wxFindMaxSize(hwndStaticMax, &rect);
    wxFindMaxSize(hwndEditValue, &rect);

    // Since we now have the absolute screen coords,
    // if there's a parent we must subtract its top left corner
    wxPOINTL point = { rect.xLeft, rect.yTop };
    if (parent) {
	wxWnd *cparent = (wxWnd *)(parent->handle);
	wx_ScreenToClient(cparent->GetClientHWND(), &point);
    }

    *x = point.x;
    *y = point.y;
    TRACE("wxSlider::GetPosition returns x=%d, y=%d", *x, *y);
}

void wxSlider::SetSize(int x, int y, int width, int height, int sizeFlags)
{
    TRACE("wxSlider::SetSize(x=%d, y=%d, w=%d, h=%d, sizeFlags=%#x)", x, y, width, height, sizeFlags);

    if (x == -1)
	x = currentX;
    if (y == -1)
	y = currentY;
	
    currentX = x;
    currentY = y;
    currentW = width;
    currentH = height;

    int cx, cy;     // slider,min,max sizes
	
    wxGetCharSize(hwndItem, &cx, &cy);
	
    if ((windowStyle & wxVERTICAL) != wxVERTICAL) {
	/*
	 * Calculate Width of control
	 */
	int control_width, control_height, edit_width, slider_width;
	float min_width, max_width, cyf;
	char buf[300];
		
	WinQueryWindowText(hwndStaticMin, 300, (PCH)buf);
	GetTextExtent(buf, &min_width, &cyf,NULL,NULL, buttonFont);
		
	WinQueryWindowText(hwndStaticMax, 300, (PCH)buf);
	GetTextExtent(buf, &max_width, &cyf,NULL,NULL, buttonFont);
	control_width = (int)(min_width + cx + cx + max_width);
		
	control_width = 0;
	control_height = WinQuerySysValue(HWND_DESKTOP, SV_CYHSCROLL);
		
	edit_width = (int)(wxMax(min_width, max_width));
	control_width += edit_width + cx;
		
	if (width > 0)
	    slider_width = width - control_width;
	else
	    slider_width = 100;
	if (slider_width < 100)
	    slider_width = 100;
		
	control_width += slider_width;
	if (height > 0)
	    control_height = height;
	else
	    control_height = (int)cyf;
		
		
	/*
	 * Now position the different sub controls
	 */
	if (HaveLabel()) {
	    int cyl = -1;
	    AdjustLabelPos(&x, &y, &control_width, &cyl);
	    control_height += cyl;
	}
		
	wx_MoveWindow(hwndEditValue, x, y, edit_width, (int)cyf, TRUE);
	x += edit_width + cx;
		
	wx_MoveWindow(hwndStaticMin, x, y, (int)min_width, cy, TRUE);
	x += (int)min_width + cx;

	MoveWindow(x, y, slider_width, cy, TRUE);
	x += slider_width + cx;

	wx_MoveWindow(hwndStaticMax, x, y, (int)max_width, cy, TRUE);
#if XXX
    } else {
	if (HaveLabel()) {
	    wxGetCharSize(hwndItem_handle, &cxs, &cys);
	    GetWindowText(GetLabelHWND(), buf, 300);
	    GetTextExtent(buf, &current_width, &cyf,NULL,NULL, labelFont);
	    MoveWindow(GetLabelHWND(), x_offset, y, (int)(current_width + 2*cxs), (int)cyf, TRUE);
	    y_offset += cys ;
	}

	float min_len;
	GetWindowText(hwndStaticMin, buf, 300);
	GetTextExtent(buf, &min_len, &cyf,NULL,NULL, buttonFont);

	float max_len;
	GetWindowText(hwndStaticMax, buf, 300);
	GetTextExtent(buf, &max_len, &cyf,NULL,NULL, buttonFont);

	int new_width = (int)(wxMax(min_len, max_len));
	MoveWindow(hwndEditValue, x_offset, y_offset, new_width, (int)cyf, TRUE);
	y_offset += (int)cyf;

	MoveWindow(hwndStaticMin, x_offset, y_offset, (int)min_len, cy, TRUE);
	y_offset += cy;

	int slider_length = (int)(height - y_offset - cy - cy);

	// Slider must have a minimum/default length
	if (slider_length < 0)
	    slider_length = 100;
		
	// Use character height as an estimate of slider width (yes, width)
	MoveWindow((HWND)ms_handle, x_offset, y_offset, cy, slider_length, TRUE);
	y_offset += slider_length;
		
	MoveWindow(hwndStaticMax, x_offset, y_offset, (int)max_len, cy, TRUE);
#endif
    }
    OnSize(width, height);
}

void wxSlider::SetRange(int minValue, int maxValue)
{
    int value = (int)WinSendMsg(hwndItem, SBM_QUERYPOS, (MPARAM)NULL, (MPARAM)NULL) + s_min;

    s_min = minValue;
    s_max = maxValue;

    char buf[40];
    sprintf(buf, "%d", s_min);
    WinSetWindowText(hwndStaticMin, (PSZ)buf);

    sprintf(buf, "%d", s_max);
    WinSetWindowText(hwndStaticMax, (PSZ)buf);

    WinSendMsg(hwndItem, SBM_SETSCROLLBAR, (MPARAM)value, MPFROM2SHORT(0, s_max - s_min));
    WinSendMsg(hwndItem, SBM_SETPOS, (MPARAM)value - s_min, (MPARAM)0);
}

int wxSlider::GetMin(void)
{
    return s_min;
}

int wxSlider::GetMax(void)
{
    return s_max;
}

void wxSlider::SetBackgroundColour(wxColour& col)
{
    TRACE("wxSlider::SetBackgroundColour(%#06lx)", (LONG)col);
    wxItem::SetBackgroundColour(col);
    if (hwndStaticMin)
	wxSetBackgroundColour(hwndStaticMin, backColour);
    if (hwndStaticMax)
	wxSetBackgroundColour(hwndStaticMax, backColour);
}

void wxSlider::SetLabelColour(wxColour& col) 
{
    TRACE("wxSlider::SetLabelColour(col=%#06lx)", (LONG)col);
    wxItem::SetLabelColour(col);
    if (hwndStaticMin)
	wxSetForegroundColour(hwndStaticMin, labelColour);
    if (hwndStaticMax)
	wxSetForegroundColour(hwndStaticMax, labelColour);
}

